# pylint: disable=no-name-in-module
# pylint: disable=too-many-lines
import re
import logging
from copy import deepcopy

import numpy as np
import pandas as pd
import pymc3 as pm

from arviz.plots import plot_posterior
from arviz.data import from_dict
from numpy.linalg import matrix_rank
from formulae import design_matrices

from .backends import PyMC3BackEnd
from .priors import Prior, PriorFactory, PriorScaler, Family
from .terms import ResponseTerm, Term, GroupSpecificTerm
from .utils import listify, extract_family_prior, link_match_family
from .version import __version__

_log = logging.getLogger("bambi")


class Model:
    """Specification of model class.

    Parameters
    ----------
    formula : str
        A model description written in model formula language.
    data : DataFrame or str
        The dataset to use. Either a pandas ``DataFrame``, or the name of the file containing
        the data, which will be passed to ``pd.read_csv()``.
    family : str or Family
        A specification of the model family (analogous to the family object in R). Either
        a string, or an instance of class ``priors.Family``. If a string is passed, a family
        with the corresponding name must be defined in the defaults loaded at ``Model``
        initialization.Valid pre-defined families are ``'gaussian'``, ``'bernoulli'``,
        ``'poisson'``, ``'gamma'``, ``'wald'``, and ``'negativebinomial'``.
        Defaults to ``'gaussian'``.
    priors : dict
        Optional specification of priors for one or more terms. A dictionary where the keys are
        the names of terms in the model, 'common' or 'group_specific' and the values are either
        instances of class ``Prior`` or ``int``, ``float``, or ``str`` that specify the
        width of the priors on a standardized scale.
    link : str
        The model link function to use. Can be either a string (must be one of the options
        defined in the current backend; typically this will include at least ``'identity'``,
        ``'logit'``, ``'inverse'``, and ``'log'``), or a callable that takes a 1D ndarray or
        theano tensor as the sole argument and returns one with the same shape.
    categorical : str or list
        The names of any variables to treat as categorical. Can be either a single variable
        name, or a list of names. If categorical is ``None``, the data type of the columns in
        the ``DataFrame`` will be used to infer handling. In cases where numeric columns are
        to be treated as categoricals (e.g., group specific factors coded as numerical IDs),
        explicitly passing variable names via this argument is recommended.
    dropna : bool
        When ``True``, rows with any missing values in either the predictors or outcome are
        automatically dropped from the dataset in a listwise manner.
    auto_scale : bool
        If ``True`` (default), priors are automatically rescaled to the data
        (to be weakly informative) any time default priors are used. Note that any priors
        explicitly set by the user will always take precedence over default priors.
    default_priors : dict or str
        An optional specification of the default priors to use for all model terms. Either a
        dictionary containing named distributions, families, and terms (see the documentation in
        ``priors.PriorFactory`` for details), or the name of a JSON file containing the same
        information.
    noncentered : bool
        If ``True`` (default), uses a non-centered parameterization for normal hyperpriors on
        grouped parameters. If ``False``, naive (centered) parameterization is used.
    taylor : int
        Order of Taylor expansion to use in approximate variance when constructing the default
        priors. Should be between 1 and 13. Lower values are less accurate, tending to undershoot
        the correct prior width, but are faster to compute and more stable. Odd-numbered values
        tend to work better. Defaults to 5 for Normal models and 1 for non-Normal models. Values
        higher than the defaults are generally not recommended as they can be unstable.
    """

    # pylint: disable=too-many-instance-attributes
    def __init__(
        self,
        formula=None,
        data=None,
        family="gaussian",
        priors=None,
        link=None,
        categorical=None,
        dropna=False,
        auto_scale=True,
        default_priors=None,
        noncentered=True,
        taylor=None,
    ):
        # attributes that are set later
        self.terms = {}
        self.dm_statistics = None  # build()
        self.built = False  # build()
        self._backend_name = None

        # build() will loop over this, calling _set_priors()
        self._added_priors = {}

        self._design = None
        self.formula = None
        self.response = None  # _add_response()
        self.family = None  # _add_response()
        self.backend = None  # _set_backend()

        self.auto_scale = auto_scale
        self.dropna = dropna
        self.taylor = taylor
        self.noncentered = noncentered

        # Read and clean data
        if isinstance(data, str):
            data = pd.read_csv(data, sep=None, engine="python")
        elif not isinstance(data, pd.DataFrame):
            raise ValueError("data must be a string with a path to a .csv or a pandas DataFrame.")

        # To avoid SettingWithCopyWarning when converting object columns to category
        data._is_copy = False

        # Object columns converted to category by default.
        obj_cols = data.select_dtypes(["object"]).columns
        data[obj_cols] = data[obj_cols].apply(lambda x: x.astype("category"))

        # Explicitly convert columns to category if desired--though this
        # can also be done within the formula using C().
        if categorical is not None:
            data = data.copy()
            cats = listify(categorical)
            data[cats] = data[cats].apply(lambda x: x.astype("category"))

        self.data = data

        # Handle priors
        if priors is None:
            priors = {}
        else:
            priors = deepcopy(priors)
        self.default_priors = PriorFactory(default_priors)

        # Obtain design matrices and related objects.
        na_action = "drop" if dropna else "error"
        if formula is not None:
            self.formula = formula
            self._design = design_matrices(formula, data, na_action, eval_env=1)
        else:
            raise ValueError("Can't instantiate a model without a model formula.")

        if self._design.response is not None:
            _family = family.name if isinstance(family, Family) else family
            priors_ = extract_family_prior(family, priors)
            if priors_ and self._design.common:
                conflicts = [name for name in priors_ if name in self._design.common.terms_info]
                if conflicts:
                    raise ValueError(
                        f"The prior name for {', '.join(conflicts)} conflicts with the name of a "
                        "parameter in the response distribution.\n"
                        "Please rename the term(s) to prevent an unexpected behaviour."
                    )
            self._add_response(self._design.response, family, link, priors_)
        else:
            raise ValueError(
                "No outcome variable is set! "
                "Please specify an outcome variable using the formula interface."
            )

        if self._design.common:
            self._add_common(self._design.common, priors)

        if self._design.group:
            self._add_group_specific(self._design.group, priors)

        # Build priors
        self._build_priors()

    def fit(
        self,
        omit_offsets=True,
        backend="pymc",
        **kwargs,
    ):
        """Fit the model using the specified backend.

        Parameters
        ----------
        omit_offsets: bool
            Omits offset terms in the ``InferenceData`` object when the model includes group
            specific effects. Defaults to ``True``.
        backend : str
            The name of the backend to use. Currently only ``'pymc'`` backend is supported.
        """

        # There's a problem if we pretend to move .build() to instantiation
        # We would have to assume the backend, which is fine now since we're using PyMC3.
        # Nevermind, prior scaling can be outside .build() bc the backend is not needed!!
        if backend is None:
            backend = "pymc" if self._backend_name is None else self._backend_name

        if not self.built or backend != self._backend_name:
            self.build(backend)
            self._backend_name = backend

        # Tell user which event is being modeled
        if self.family.name == "bernoulli":
            _log.info(
                "Modeling the probability that %s==%s",
                self.response.name,
                str(self.response.success_event),
            )

        return self.backend.run(omit_offsets=omit_offsets, **kwargs)

    def build(self, backend="pymc"):
        """Set up the model for sampling/fitting.

        Performs any steps that require access to all model terms (e.g., scaling priors
        on each term), then calls the backend's ``build()`` method.

        Parameters
        ----------
        backend : str
            The name of the backend to use for model fitting.
            Currently only ``'pymc'`` is supported.
        """

        # Check for backend
        if backend is None:
            if self._backend_name is None:
                raise ValueError(
                    "No backend was passed or set in the Model; did you forget to call fit()?"
                )
            backend = self._backend_name

        self._set_backend(backend)
        self.backend.build(self)
        self.built = True

    def set_priors(self, priors=None, common=None, group_specific=None, match_derived_names=True):
        """Set priors for one or more existing terms.

        Parameters
        ----------
        priors : dict
            Dictionary of priors to update. Keys are names of terms to update; values are the new
            priors (either a ``Prior`` instance, or an int or float that scales the default priors).
            Note that a tuple can be passed as the key, in which case the same prior will be applied
            to all terms named in the tuple.
        common : Prior, int, float or str
            A prior specification to apply to all common terms included in the model.
        group_specific : Prior, int, float or str
            A prior specification to apply to all group specific terms included in the model.
        match_derived_names : bool
            If ``True``, the specified prior(s) will be applied not only to terms that match the
            keyword exactly, but to the levels of group specific effects that were derived from the
            original specification with the passed name. For example,
            ``priors={'condition|subject':0.5}`` would apply the prior to the terms with names
            ``'1|subject'``, ``'condition[T.1]|subject'``, and so on. If ``False``, an exact
            match is required for the prior to be applied.
        """
        # save arguments to pass to _set_priors() at build time
        kwargs = dict(
            zip(
                ["priors", "common", "group_specific", "match_derived_names"],
                [priors, common, group_specific, match_derived_names],
            )
        )
        self._added_priors.update(kwargs)
        # After updating, we need to rebuild priors.
        # There is redundancy here, so there's place for performance improvements.
        self._build_priors()
        self.built = False

    def _set_backend(self, backend):
        backend = backend.lower()
        if backend.startswith("pymc"):
            self.backend = PyMC3BackEnd()
        else:
            raise ValueError("At the moment, only the PyMC3 backend is supported.")

        self._backend_name = backend

    def _build_priors(self):
        """Carry out all operations related to the construction and/or scaling of priors."""
        # set custom priors
        self._set_priors(**self._added_priors)

        # prepare all priors
        for name, term in self.terms.items():
            type_ = (
                "intercept"
                if name == "Intercept"
                else "group_specific"
                if self.terms[name].group_specific
                else "common"
            )
            term.prior = self._prepare_prior(term.prior, type_)

        # Only compute the mean stats if there are multiple terms in the model
        terms = [t for t in self.common_terms.values() if t.name != "Intercept"]

        if len(self.common_terms) > 1:
            x_matrix = [pd.DataFrame(x.data, columns=x.levels) for x in terms]
            x_matrix = pd.concat(x_matrix, axis=1)
            self.dm_statistics = {"mean_x": x_matrix.mean(axis=0)}

        # throw informative error message if any categorical predictors have 1 category
        num_cats = [x.data.size for x in self.common_terms.values()]
        if any(np.array(num_cats) == 0):
            raise ValueError("At least one categorical predictor contains only 1 category!")

        # only set priors if there is at least one term in the model
        if self.terms:
            # Get and scale default priors if none are defined yet
            if self.taylor is not None:
                taylor = self.taylor
            else:
                taylor = 5 if self.family.name == "gaussian" else 1
            scaler = PriorScaler(self, taylor=taylor)
            scaler.scale()

    def _set_priors(self, priors=None, common=None, group_specific=None, match_derived_names=True):
        """Internal version of ``set_priors()``, with same arguments.

        Runs during ``Model._build_priors()``.
        """
        targets = {}

        if common is not None:
            targets.update({name: common for name in self.common_terms.keys()})

        if group_specific is not None:
            targets.update({name: group_specific for name in self.group_specific_terms.keys()})

        if priors is not None:
            # Update priors related to nuisance parameters of response distribution
            priors_ = extract_family_prior(self.family, priors)
            if priors_:
                # Remove keys passed to the response.
                for key in priors_:
                    priors.pop(key)
                self.response.prior.args.update(priors_)

            # Prepare priors for explanatory terms.
            for k, prior in priors.items():
                for name in listify(k):
                    term_names = list(self.terms.keys())
                    msg = f"No terms in model match {name}."
                    if name not in term_names:
                        terms = self._match_derived_terms(name)
                        if not match_derived_names or terms is None:
                            raise ValueError(msg)
                        for term in terms:
                            targets[term.name] = prior
                    else:
                        targets[name] = prior

        # Set priors for explanatory terms.
        for name, prior in targets.items():
            self.terms[name].prior = prior

    def _prepare_prior(self, prior, _type):
        """Helper function to correctly set default priors, auto scaling, etc.

        Parameters
        ----------
        prior : Prior object, or float, or None.
        _type : string
            accepted values are: ``'intercept'``, ``'common'``, or ``'group_specific'``.
        """
        if prior is None and not self.auto_scale:
            prior = self.default_priors.get(term=_type + "_flat")

        if isinstance(prior, Prior):
            prior._auto_scale = False  # pylint: disable=protected-access
        else:
            _scale = prior
            prior = self.default_priors.get(term=_type)
            prior.scale = _scale
            if prior.scale is not None:
                prior._auto_scale = False  # pylint: disable=protected-access
        return prior

    def _add_response(self, response, family="gaussian", link=None, priors=None):
        """Add a response (or outcome/dependent) variable to the model.

        Parameters
        ----------
        response : formulae.ResponseVector
            An instance of ``formulae.ResponseVector`` as returned by
            ``formulae.design_matrices()``.
        family : str or Family
            A specification of the model family (analogous to the family object in R). Either a
            string, or an instance of class ``priors.Family``. If a string is passed, a family with
            the corresponding name must be defined in the defaults loaded at Model initialization.
            Valid pre-defined families are ``'gaussian'``, ``'bernoulli'``, ``'poisson'``,
            ``'gamma'``, ``'wald'``, and ``'negativebinomial'``. Defaults to ``'gaussian'``.
        link : str
            The model link function to use. Can be either a string (must be one of the options
            defined in the current backend; typically this will include at least ``'identity'``,
            ``'logit'``, ``'inverse'``, and ``'log'``), or a callable that takes a 1D ndarray or
            theano tensor as the sole argument and returns one with the same shape.
        priors : dict
            Optional dictionary with specification of priors for the parameters in the family of
            the response. Keys are names of other parameters than the mean in the family
            (i.e. they cannot be equal to family.parent) and values can be an instance of class
            ``Prior``, a numeric value, or a string describing the width. In the numeric case,
            the distribution specified in the defaults will be used, and the passed value will be
            used to scale the appropriate variance parameter. For strings (e.g., ``'wide'``,
            ``'narrow'``, ``'medium'``, or ``'superwide'``), predefined values will be used.
        """
        if isinstance(family, str):
            family = self.default_priors.get(family=family)
            if family.name == "gamma":
                # Drop 'beta' param. Should be handled better in the future.
                family.prior.args.pop("beta")
        elif not isinstance(family, Family):
            raise ValueError("family must be a string or a Family object.")

        self.family = family

        # Override family's link if another is explicitly passed
        if link is not None:
            if link_match_family(link, family.name):
                self.family._set_link(link)  # pylint: disable=protected-access
            else:
                raise ValueError(f"Link {link} cannot be used with family {family.name}")

        prior = self.family.prior

        # not None when user passes priors for nuisance parameters, either for built-in familes or
        # for custom families.
        if priors is not None:
            prior.args.update(priors)

        if self.family.name == "gaussian":
            if priors is None:
                prior.update(
                    sigma=Prior("HalfStudentT", nu=4, sigma=np.std(response.design_vector))
                )

        if response.refclass is not None and self.family.name != "bernoulli":
            raise ValueError("Index notation for response only available for 'bernoulli' family")

        self.response = ResponseTerm(response, prior, self.family.name)
        self.built = False

    def _add_common(self, common, priors):
        """Add common (or fixed) terms to the model.

        Parameters
        ----------
        common : formulae.CommonEffectsMatrix
            Representation of the design matrix for the common effects of a model. It contains all
            the necessary information to build the ``Term`` objects associated with each common
            term in the model.
        priors : dict
            Optional specification of priors for one or more terms. A dictionary where the keys are
            any of the names of the common terms in the model or 'common' and the values are either
            instances of class ``Prior`` or ``int``, ``float``, or ``str`` that specify the width
            of the priors on a standardized scale.
        """
        if matrix_rank(common.design_matrix) < common.design_matrix.shape[1]:
            raise ValueError(
                "Design matrix for common effects is not full-rank. "
                "Bambi does not support sparse settings yet."
            )

        for name, term in common.terms_info.items():
            data = common[name]
            prior = priors.pop(name, priors.get("common", None))
            if isinstance(prior, Prior):
                any_hyperprior = any(isinstance(x, Prior) for x in prior.args.values())
                if any_hyperprior:
                    raise ValueError(
                        f"Trying to set hyperprior on '{name}'. "
                        "Can't set a hyperprior on common effects."
                    )
            self.terms[name] = Term(name, term, data, prior)

    def _add_group_specific(self, group, priors):
        """Add group-specific (or random) terms to the model.

        Parameters
        ----------
        group : formulae.GroupEffectsMatrix
            Representation of the design matrix for the group specific effects of a model. It
            contains all the necessary information to build the ``GroupSpecificTerm`` objects
            associated with each group-specific term in the model.
        priors : dict
            Optional specification of priors for one or more terms. A dictionary where the keys are
            any of the names of the group-specific terms in the model or 'group_specific' and the
            values are either instances of class ``Prior`` or ``int``, ``float``, or ``str``
            that specify the width of the priors on a standardized scale.
        """
        for name, term in group.terms_info.items():
            data = group[name]
            prior = priors.pop(name, priors.get("group_specific", None))
            self.terms[name] = GroupSpecificTerm(name, term, data, prior)

    def _match_derived_terms(self, name):
        """Return all (group_specific) terms whose named are derived from the specified string.

        For example, ``'condition|subject'`` should match the terms with names ``'1|subject'``,
        ``'condition[T.1]|subject'``, and so on.
        Only works for strings with grouping operator ``('|')``.
        """
        if "|" not in name:
            return None

        patt = r"^([01]+)*[\s\+]*([^\|]+)*\|(.*)"
        intcpt, pred, grpr = re.search(patt, name).groups()
        intcpt = f"1|{grpr}"
        if not pred:
            return [self.terms[intcpt]] if intcpt in self.terms else None

        source = f"{pred}|{grpr}"
        found = [
            t
            for (n, t) in self.terms.items()
            if n == intcpt or re.sub(r"(\[.*?\])", "", n) == source
        ]
        # If only the intercept matches, return None, because we want to err
        # on the side of caution and not consider '1|subject' to be a match for
        # 'condition|subject' if no slopes are found (e.g., the intercept could
        # have been set by some other specification like 'gender|subject').
        return found if found and (len(found) > 1 or found[0].name != intcpt) else None

    def plot_priors(
        self,
        draws=5000,
        var_names=None,
        random_seed=None,
        figsize=None,
        textsize=None,
        hdi_prob=None,
        round_to=2,
        point_estimate="mean",
        kind="kde",
        bins=None,
        omit_offsets=True,
        omit_group_specific=True,
        ax=None,
    ):
        """
        Samples from the prior distribution and plots its marginals.

        Parameters
        ----------
        draws : int
            Number of draws to sample from the prior predictive distribution. Defaults to 5000.
        var_names : str or list
            A list of names of variables for which to compute the posterior predictive
            distribution. Defaults to both observed and unobserved RVs.
        random_seed : int
            Seed for the random number generator.
        figsize: tuple
            Figure size. If ``None`` it will be defined automatically.
        textsize: float
            Text size scaling factor for labels, titles and lines. If ``None`` it will be
            autoscaled based on ``figsize``.
        hdi_prob: float
            Plots highest density interval for chosen percentage of density.
            Use ``'hide'`` to hide the highest density interval. Defaults to 0.94.
        round_to: int
            Controls formatting of floats. Defaults to 2 or the integer part, whichever is bigger.
        point_estimate: str
            Plot point estimate per variable. Values should be ``'mean'``, ``'median'``, ``'mode'``
             or ``None``. Defaults to ``'auto'`` i.e. it falls back to default set in
             ArviZ's rcParams.
        kind: str
            Type of plot to display (``'kde'`` or ``'hist'``) For discrete variables this argument
            is ignored and a histogram is always used.
        bins: integer or sequence or 'auto'
            Controls the number of bins, accepts the same keywords ``matplotlib.hist()`` does.
            Only works if ``kind == hist``. If ``None`` (default) it will use ``auto`` for
            continuous variables and ``range(xmin, xmax + 1)`` for discrete variables.
        omit_offsets: bool
            Whether to omit offset terms in the plot. Defaults to ``True``.
        omit_group_specific: bool
            Whether to omit group specific effects in the plot. Defaults to ``True``.
        ax: numpy array-like of matplotlib axes or bokeh figures
            A 2D array of locations into which to plot the densities. If not supplied, ArviZ will
            create its own array of plot areas (and return it).
        **kwargs
            Passed as-is to ``plt.hist()`` or ``plt.plot()`` function depending on the value of
            ``kind``.

        Returns
        -------
        axes: matplotlib axes or bokeh figures
        """
        if not self.built:
            raise ValueError(
                "Cannot plot priors until model is built!! "
                "Call .build() to build the model or .fit() to build and sample from the posterior."
            )

        unobserved_rvs_names = []
        flat_rvs = []
        for unobserved in self.backend.model.unobserved_RVs:
            if "Flat" in unobserved.__str__():
                flat_rvs.append(unobserved.name)
            else:
                unobserved_rvs_names.append(unobserved.name)
        if var_names is None:
            var_names = pm.util.get_default_varnames(
                unobserved_rvs_names, include_transformed=False
            )
        else:
            flat_rvs = [fv for fv in flat_rvs if fv in var_names]
            var_names = [vn for vn in var_names if vn not in flat_rvs]

        if flat_rvs:
            _log.info(
                "Variables %s have flat priors, and hence they are not plotted", ", ".join(flat_rvs)
            )

        if omit_offsets:
            omitted = [f"{rt}_offset" for rt in self.group_specific_terms]
            var_names = [vn for vn in var_names if vn not in omitted]

        if omit_group_specific:
            omitted = list(self.group_specific_terms)
            var_names = [vn for vn in var_names if vn not in omitted]

        axes = None
        if var_names:
            pps = self.prior_predictive(draws=draws, var_names=var_names, random_seed=random_seed)

            axes = plot_posterior(
                pps,
                group="prior",
                figsize=figsize,
                textsize=textsize,
                hdi_prob=hdi_prob,
                round_to=round_to,
                point_estimate=point_estimate,
                kind=kind,
                bins=bins,
                ax=ax,
            )
        return axes

    def prior_predictive(self, draws=500, var_names=None, omit_offsets=True, random_seed=None):
        """
        Generate samples from the prior predictive distribution.

        Parameters
        ----------
        draws : int
            Number of draws to sample from the prior predictive distribution. Defaults to 500.
        var_names : str or list
            A list of names of variables for which to compute the posterior predictive
            distribution. Defaults to both observed and unobserved RVs.
        random_seed : int
            Seed for the random number generator.

        Returns
        -------
        InferenceData
            ``InferenceData`` object with the groups `prior`, ``prior_predictive`` and
            ``observed_data``.
        """
        if var_names is None:
            variables = self.backend.model.unobserved_RVs + self.backend.model.observed_RVs
            variables_names = [v.name for v in variables]
            var_names = pm.util.get_default_varnames(variables_names, include_transformed=False)

        if omit_offsets:
            offset_vars = [f"{rt}_offset" for rt in self.group_specific_terms]
            var_names = [vn for vn in var_names if vn not in offset_vars]

        pps_ = pm.sample_prior_predictive(
            samples=draws, var_names=var_names, model=self.backend.model, random_seed=random_seed
        )
        # pps_ keys are not in the same order as `var_names` because `var_names` is converted
        # to set within pm.sample_prior_predictive()
        pps = {name: pps_[name] for name in var_names}

        response_name = self.response.name

        if response_name in pps:
            prior_predictive = {response_name: np.moveaxis(pps.pop(response_name), 2, 0)}
            observed_data = {response_name: self.response.data.squeeze()}
        else:
            prior_predictive = {}
            observed_data = {}

        prior = {k: v[np.newaxis] for k, v in pps.items()}

        idata = from_dict(
            prior_predictive=prior_predictive,
            prior=prior,
            observed_data=observed_data,
            coords=self.backend.model.coords,  # new line
            attrs={
                "inference_library": self.backend.name,
                "inference_library_version": self.backend.name,
                "modeling_interface": "bambi",
                "modeling_interface_version": __version__,
            },
        )

        return idata

    def posterior_predictive(
        self, idata, draws=500, var_names=None, inplace=True, random_seed=None
    ):
        """
        Generate samples from the posterior predictive distribution.

        Parameters
        ----------
        idata : InfereceData
            ``InfereceData`` with samples from the posterior distribution.
        draws : int
            Number of draws to sample from the prior predictive distribution. Defaults to 500.
        var_names : str or list
            A list of names of variables for which to compute the posterior predictive
            distribution. Defaults to both observed and unobserved RVs.
        inplace : bool
            If ``True`` it will add a ``posterior_predictive`` group to idata, otherwise it will
            return a copy of idata with the added group. If ``True`` and idata already have a
            ``posterior_predictive`` group it will be overwritten.
        random_seed : int
            Seed for the random number generator.

        Returns
        -------
        None or InferenceData
            When ``inplace=True`` add ``posterior_predictive`` group to idata and return
            ``None``. Otherwise a copy of idata with a ``posterior_predictive`` group.

        """
        if var_names is None:
            variables = self.backend.model.observed_RVs
            variables_names = [v.name for v in variables]
            var_names = pm.util.get_default_varnames(variables_names, include_transformed=False)

        pps = pm.sample_posterior_predictive(
            trace=idata,
            samples=draws,
            var_names=var_names,
            model=self.backend.model,
            random_seed=random_seed,
        )

        if not inplace:
            idata = deepcopy(idata)
        if "posterior_predictive" in idata:
            del idata.posterior_predictive

        idata.add_groups(
            {"posterior_predictive": {k: v.squeeze()[np.newaxis] for k, v in pps.items()}}
        )

        getattr(idata, "posterior_predictive").attrs["modeling_interface"] = "bambi"
        getattr(idata, "posterior_predictive").attrs["modeling_interface_version"] = __version__
        if inplace:
            return None
        else:
            return idata

    def graph(self, formatting="plain", name=None, figsize=None, dpi=300, fmt="png"):
        """
        Produce a graphviz Digraph from a Bambi model.

        Requires graphviz, which may be installed most easily with
            ``conda install -c conda-forge python-graphviz``

        Alternatively, you may install the ``graphviz`` binaries yourself, and then
        ``pip install graphviz`` to get the python bindings.
        See http://graphviz.readthedocs.io/en/stable/manual.html for more information.

        Parameters
        ----------
        formatting : str
            One of ``'plain'`` or ``'plain_with_params'``. Defaults to ``'plain'``.
        name : str
            Name of the figure to save. Defaults to None, no figure is saved.
        figsize : tuple
            Maximum width and height of figure in inches. Defaults to None, the figure size is set
            automatically. If defined and the drawing is larger than the given size, the drawing is
            uniformly scaled down so that it fits within the given size.  Only works if ``name``
            is not None.
        dpi : int
            Point per inch of the figure to save.
            Defaults to 300. Only works if ``name`` is not None.
        fmt : str
            Format of the figure to save.
            Defaults to ``'png'``. Only works if ``name`` is not None.
        """
        if self.backend is None:
            raise ValueError("The model is empty, please define a Bambi model")

        graphviz = pm.model_to_graphviz(model=self.backend.model, formatting=formatting)

        width, height = (None, None) if figsize is None else figsize

        if name is not None:
            graphviz_ = graphviz.copy()
            graphviz_.graph_attr.update(size=f"{width},{height}!")
            graphviz_.graph_attr.update(dpi=str(dpi))
            graphviz_.render(filename=name, format=fmt, cleanup=True)

        return graphviz

    def _get_pymc_coords(self):
        # categorical attribute is important because of this coordinates stuff
        common_terms = {
            k + "_dim_0": v.cleaned_levels for k, v in self.common_terms.items() if v.categorical
        }
        # Include all group specific terms
        group_specific_terms = {
            k + "_dim_0": v.cleaned_levels for k, v in self.group_specific_terms.items()
        }
        return {**common_terms, **group_specific_terms}

    def __str__(self):
        priors = [f"  {term.name} ~ {term.prior}" for term in self.terms.values()]
        # Priors for nuisance parameters, i.e., standard deviation in normal linear model
        priors_extra_params = [
            f"  {k} ~ {v}"
            for k, v in self.family.prior.args.items()
            if k not in ["observed", self.family.parent]
        ]
        priors += priors_extra_params
        str_list = [
            f"Formula: {self.formula}",
            f"Family name: {self.family.name.capitalize()}",
            f"Link: {self.family.link}",
            f"Observations: {self.response.data.shape[0]}",
            "Priors:",
            "\n".join(priors),
        ]
        if self.backend and self.backend.fit:
            extra_foot = "------\n"
            extra_foot += "* To see a plot of the priors call the .plot_priors() method.\n"
            extra_foot += "* To see a summary or plot of the posterior pass the object returned"
            extra_foot += " by .fit() to az.summary() or az.plot_trace()\n"
            str_list += [extra_foot]

        return "\n".join(str_list)

    def __repr__(self):
        return self.__str__()

    @property
    def term_names(self):
        """Return names of all terms in order of addition to model."""
        return list(self.terms.keys())

    @property
    def common_terms(self):
        """Return dict of all and only common effects in model."""
        return {k: v for (k, v) in self.terms.items() if not v.group_specific}

    @property
    def group_specific_terms(self):
        """Return dict of all and only group specific effects in model."""
        return {k: v for (k, v) in self.terms.items() if v.group_specific}



import logging

from .models import Model
from .priors import Prior, Family
from .backends import PyMC3BackEnd
from .version import __version__


__all__ = ["Model", "Prior", "Family", "PyMC3BackEnd"]

_log = logging.getLogger("bambi")

if not logging.root.handlers:
    _log.setLevel(logging.INFO)
    if len(_log.handlers) == 0:
        handler = logging.StreamHandler()
        _log.addHandler(handler)



# pylint: disable=no-name-in-module
import json
import re
import sys

from copy import deepcopy
from os.path import dirname, join

import numpy as np
import pandas as pd
from scipy.special import hyp2f1
from statsmodels.genmod import families as genmod_families
from statsmodels.genmod.generalized_linear_model import GLM
from statsmodels.tools.sm_exceptions import PerfectSeparationError

genmod_links = genmod_families.links


class Family:
    """A specification of model family.

    Parameters
    ----------
    name : str
        Family name.
    prior : Prior
        A ``Prior`` instance specifying the model likelihood prior.
    link : str
        The name of the link function transforming the linear model prediction to a parameter of
        the likelihood.
    parent : str
        The name of the prior parameter to set to the link-transformed predicted outcome
        (e.g., ``'mu'``, ``'p'``, etc.).
    """

    def __init__(self, name, prior, link, parent):
        self.smlink = None
        fams = {
            "bernoulli": genmod_families.Binomial,
            "gamma": genmod_families.Gamma,
            "gaussian": genmod_families.Gaussian,
            "wald": genmod_families.InverseGaussian,
            "negativebinomial": genmod_families.NegativeBinomial,
            "poisson": genmod_families.Poisson,
            "t": None,  # not implemented in statsmodels
        }
        self.name = name
        self.prior = prior
        self.link = link
        self.parent = parent
        self.smfamily = fams.get(name, None)

    def _set_link(self, link):
        """Set new link function.
        It updates both ``self.link`` (a string consumed by the backend) and ``self.smlink``
        (the link instance for the statsmodel family)
        """
        links = {
            "identity": genmod_links.identity(),
            "logit": genmod_links.logit(),
            "probit": genmod_links.probit(),
            "cloglog": genmod_links.cloglog(),
            "inverse": genmod_links.inverse_power(),
            "inverse_squared": genmod_links.inverse_squared(),
            "log": genmod_links.log(),
        }
        self.link = link
        if link in links:
            self.smlink = links[link]
        else:
            raise ValueError("Link name is not supported.")

    def __str__(self):
        if self.smfamily is None:
            str_ = "No family set"
        else:
            priors_str = ",\n  ".join(
                [
                    f"{k}: {np.round_(v, 8)}" if not isinstance(v, Prior) else f"{k}: {v}"
                    for k, v in self.prior.args.items()
                    if k not in ["observed", self.parent]
                ]
            )
            str_ = f"Response distribution: {self.smfamily.__name__}\n"
            str_ += f"Link: {self.link}\n"
            str_ += f"Priors:\n  {priors_str}"
        return str_

    def __repr__(self):
        return self.__str__()


class Prior:
    """Abstract specification of a term prior.

    Parameters
    ----------
    name : str
        Name of prior distribution (e.g., ``'Normal'``, ``'Bernoulli'``, etc.)
    kwargs : dict
        Optional keywords specifying the parameters of the named distribution.
    """

    def __init__(self, name, scale=None, **kwargs):
        self.name = name
        self._auto_scale = True
        self.scale = scale
        self.args = {}
        self.update(**kwargs)

    def update(self, **kwargs):
        """Update the model arguments with additional arguments.

        Parameters
        ----------
            kwargs : dict
                Optional keyword arguments to add to prior args.
        """
        # Backends expect numpy arrays, so make sure all numeric values are represented as such.
        kwargs_ = {}
        for key, val in kwargs.items():
            if isinstance(val, (int, float)):
                val = np.array(val)
            elif isinstance(val, np.ndarray):
                val = val.squeeze()
            kwargs_[key] = val
        self.args.update(kwargs_)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        else:
            return self.__dict__ == other.__dict__

    def __str__(self):
        args = ", ".join(
            [
                f"{k}: {np.round_(v, 8)}" if not isinstance(v, Prior) else f"{k}: {v}"
                for k, v in self.args.items()
            ]
        )
        return f"{self.name}({args})"

    def __repr__(self):
        return self.__str__()


class PriorFactory:
    """An object that supports specification and easy retrieval of default priors.

    Parameters
    ----------
    defaults : str or dict
        Optional base configuration containing default priors for distribution, families, and term
        types. If a string, the name of a JSON file containing the config. If a dict, must contain
        keys for ``'dists'``, ``'terms'``, and ``'families'``; see the built-in JSON configuration
        for an example. If ``None``, a built-in set of priors will be used as defaults.
    dists : dict
        Optional specification of named distributions to use as priors. Each key gives the name of
        a newly defined distribution; values are two-element lists, where the first element is the
        name of the built-in distribution to use (``'Normal'``, ``'Cauchy',`` etc.), and the second
        element is a dictionary of parameters on that distribution
        (e.g., ``{'mu': 0, 'sigma': 10}``). Priors can be nested to arbitrary depths by replacing
        any parameter with another prior specification.
    terms : dict
        Optional specification of default priors for different model term types. Valid keys are
        ``'intercept'``, ``'common'``, or ``'group_specific'``. Values are either strings preprended
        by a #, in which case they are interpreted as pointers to distributions named in
        the dists dictionary, or key -> value specifications in the same format as elements in
        the dists dictionary.
    families : dict
        Optional specification of default priors for named family objects. Keys are family names,
        and values are dicts containing mandatory keys for ``'dist'``, ``'link'``, and ``'parent'``.

    Examples
    --------
        >>> dists = {'my_dist': ['Normal', {'mu': 10, 'sigma': 1000}]}
        >>> pf = PriorFactory(dists=dists)

        >>> families = {'normalish': {'dist': ['normal', {sigma: '#my_dist'}],
        >>>             link:'identity', parent: 'mu'}}
        >>> pf = PriorFactory(dists=dists, families=families)
    """

    def __init__(self, defaults=None, dists=None, terms=None, families=None):
        if defaults is None:
            defaults = join(dirname(__file__), "config", "priors.json")

        if isinstance(defaults, str):
            with open(defaults, "r") as ofile:
                defaults = json.load(ofile)

        # Just in case the user plans to use the same defaults elsewhere
        defaults = deepcopy(defaults)

        if isinstance(dists, dict):
            defaults["dists"].update(dists)

        if isinstance(terms, dict):
            defaults["terms"].update(terms)

        if isinstance(families, dict):
            defaults["families"].update(families)

        self.dists = defaults["dists"]
        self.terms = defaults["terms"]
        self.families = defaults["families"]

    def _get_prior(self, spec, **kwargs):
        if isinstance(spec, str):
            spec = re.sub(r"^\#", "", spec)
            return self._get_prior(self.dists[spec])
        elif isinstance(spec, (list, tuple)):
            name, args = spec
            if name.startswith("#"):
                name = re.sub(r"^\#", "", name)
                prior = self._get_prior(self.dists[name])
            else:
                prior = Prior(name, **kwargs)
            args = {k: self._get_prior(v) for (k, v) in args.items()}
            prior.update(**args)
            return prior
        else:
            return spec

    # pylint: disable=inconsistent-return-statements
    def get(self, dist=None, term=None, family=None):
        """Retrieve default prior for a named distribution, term type, or family.

        Parameters
        ----------
        dist : str
            Name of desired distribution. Note that the name is the key in the defaults dictionary,
            not the name of the ``Distribution`` object used to construct the prior.
        term : str
            The type of term family to retrieve defaults for. Must be one of ``'intercept'``,
            ``'common'``, or ``'group_specific'``.
        family : str
            The name of the ``Family`` to retrieve. Must be a value defined internally.
            In the default config, this is one of  ``'gaussian'``, ``'bernoulli'``, ``'poisson'``,
            ``'gama'``, ``'wald'``, or ``'negativebinomial'``.
        """

        if dist is not None:
            if dist not in self.dists:
                raise ValueError(f"{dist} is not a valid distribution name.")
            return self._get_prior(self.dists[dist])
        elif term is not None:
            if term not in self.terms:
                raise ValueError(f"{term} is not a valid term type.")
            return self._get_prior(self.terms[term])
        elif family is not None:
            if family not in self.families:
                raise ValueError(f"{family} is not a valid family name.")
            _f = self.families[family]
            prior = self._get_prior(_f["dist"])
            return Family(family, prior, _f["link"], _f["parent"])


class PriorScaler:
    # Default is 'wide'. The wide prior sigma is sqrt(1/3) = .577 on the partial
    # corr scale, which is the sigma of a flat prior over [-1,1].
    names = {"narrow": 0.2, "medium": 0.4, "wide": 3 ** -0.5, "superwide": 0.8}

    def __init__(self, model, taylor):
        self.model = model
        self.stats = model.dm_statistics if hasattr(model, "dm_statistics") else None
        self.dm = pd.DataFrame(
            {
                lev: term.data[:, i]
                for term in model.common_terms.values()
                for i, lev in enumerate(term.levels)
            }
        )
        self.priors = {}
        self.mle = None
        self.taylor = taylor
        with open(join(dirname(__file__), "config", "derivs.txt"), "r") as file:
            self.deriv = [next(file).strip("\n") for x in range(taylor + 1)]

    def _get_slope_stats(self, exog, predictor, sigma_corr, full_mod=None, points=4):
        """
        Parameters
        ----------
            full_mod : statsmodels.genmod.generalized_linear_model.GLM
                Statsmodels GLM to replace MLE model. For when ``'predictor'`` is not in the common
                part of the model.
            points : int
                Number of points to use for LL approximation.
        """
        if full_mod is None:
            full_mod = self.mle
        # figure out which column of exog to drop for the null model
        keeps = [
            i
            for i, x in enumerate(list(exog.columns))
            if not np.array_equal(predictor, exog[x].values.flatten())
        ]
        i = [x for x in range(exog.shape[1]) if x not in keeps][0]

        # get log-likelihood values from beta=0 to beta=MLE
        values = np.linspace(0.0, full_mod.params[i], points)
        # if there are multiple predictors, use statsmodels to optimize the LL
        if keeps:
            null = [
                GLM(
                    endog=self.model.response.data,
                    exog=exog,
                    family=self.model.family.smfamily(self.model.family.smlink),
                ).fit_constrained(str(exog.columns[i]) + "=" + str(val))
                for val in values[:-1]
            ]
            null = np.append(null, full_mod)
            log_likelihood = np.array([x.llf for x in null])
        # if just a single predictor, use statsmodels to evaluate the LL
        else:
            null = [
                self.model.family.smfamily(self.model.family.smlink).loglike(
                    np.squeeze(self.model.response.data), val * predictor
                )
                for val in values[:-1]
            ]
            log_likelihood = np.append(null, full_mod.llf)

        # compute params of quartic approximatino to log-likelihood
        # c: intercept, d: shift parameter
        # a: quartic coefficient, b: quadratic coefficient

        intercept, shift_parameter = log_likelihood[-1], -(full_mod.params[i].item())
        X = np.array([(values + shift_parameter) ** 4, (values + shift_parameter) ** 2]).T
        coef_a, coef_b = np.squeeze(
            np.linalg.multi_dot(
                [np.linalg.inv(np.dot(X.T, X)), X.T, (log_likelihood[:, None] - intercept)]
            )
        )

        # m, v: mean and variance of beta distribution of correlations
        # p, q: corresponding shape parameters of beta distribution
        mean = 0.5
        variance = sigma_corr ** 2 / 4
        p = mean * (mean * (1 - mean) / variance - 1)
        q = (1 - mean) * (mean * (1 - mean) / variance - 1)

        # function to return central moments of rescaled beta distribution
        def moment(k):
            return (2 * p / (p + q)) ** k * hyp2f1(p, -k, p + q, (p + q) / p)

        # evaluate the derivatives of beta = f(correlation).
        # dict 'point' gives points about which to Taylor expand. We want to
        # expand about the mean (generally 0), but some of the derivatives
        # do not exist at 0. Evaluating at a point very close to 0 (e.g., .001)
        # generally gives good results, but the higher order the expansion, the
        # further from 0 we need to evaluate the derivatives, or they blow up.
        point = dict(zip(range(1, 14), 2 ** np.linspace(-1, 5, 13) / 100))
        vals = dict(a=coef_a, b=coef_b, n=len(self.model.response.data), r=point[self.taylor])
        _deriv = [eval(x, globals(), vals) for x in self.deriv]  # pylint: disable=eval-used

        # compute and return the approximate sigma
        def term(i, j):
            return (
                1
                / np.math.factorial(i)
                * 1
                / np.math.factorial(j)
                * _deriv[i]
                * _deriv[j]
                * (moment(i + j) - moment(i) * moment(j))
            )

        terms = [term(i, j) for i in range(1, self.taylor + 1) for j in range(1, self.taylor + 1)]
        return np.array(terms).sum() ** 0.5

    def _get_intercept_stats(self, add_slopes=True):
        # start with mean and variance of Y on the link scale
        mod = GLM(
            endog=self.model.response.data,
            exog=np.repeat(1, len(self.model.response.data)),
            family=self.model.family.smfamily(self.model.family.smlink),
            missing="drop" if self.model.dropna else "none",
        ).fit()
        mu = mod.params
        # multiply SE by sqrt(N) to turn it into (approx.) sigma(Y) on link scale
        sigma = (mod.cov_params()[0] * len(mod.mu)) ** 0.5

        # modify mu and sigma based on means and sigmas of slope priors.
        if len(self.model.common_terms) > 1 and add_slopes:
            means = np.array([x["mu"] for x in self.priors.values()])
            sigmas = np.array([x["sigma"] for x in self.priors.values()])
            # add to intercept prior
            index = list(self.priors.keys())
            mu -= np.dot(means, self.stats["mean_x"][index])
            sigma = (sigma ** 2 + np.dot(sigmas ** 2, self.stats["mean_x"][index] ** 2)) ** 0.5

        return mu, sigma

    def _scale_common(self, term):

        # these defaults are only defined for Normal priors
        if term.prior.name != "Normal":
            return

        mu = []
        sigma = []
        sigma_corr = term.prior.scale
        for pred in term.data.T:
            mu += [0]
            sigma += [self._get_slope_stats(exog=self.dm, predictor=pred, sigma_corr=sigma_corr)]

        # save and set prior
        for i, lev in enumerate(term.levels):
            self.priors.update({lev: {"mu": mu[i], "sigma": sigma[i]}})
        term.prior.update(mu=np.array(mu), sigma=np.array(sigma))

    def _scale_intercept(self, term):

        # default priors are only defined for Normal priors
        if term.prior.name != "Normal":
            return

        # get prior mean and sigma for common intercept
        mu, sigma = self._get_intercept_stats()

        # save and set prior
        term.prior.update(mu=mu, sigma=sigma)

    def _scale_group_specific(self, term):

        # these default priors are only defined for HalfNormal priors
        if term.prior.args["sigma"].name != "HalfNormal":
            return

        sigma_corr = term.prior.scale

        # recreate the corresponding common effect data
        fix_data = term.data.sum(axis=1)

        # handle intercepts and cell means
        if term.constant:
            _, sigma = self._get_intercept_stats()
            sigma *= sigma_corr
        # handle slopes
        else:
            exists = [
                x
                for x in self.dm.columns  # pylint: disable=not-an-iterable
                if np.array_equal(fix_data, self.dm[x].values)
            ]
            # handle case where there IS a corresponding common effect
            if exists and exists[0] in self.priors.keys():
                sigma = self.priors[exists[0]]["sigma"]
            # handle case where there IS NOT a corresponding common effect
            else:
                # the usual case: add the group specific effect data as a common effect
                # in the design matrix
                if not exists:
                    fix_dataframe = pd.DataFrame(fix_data)
                    # things break if column names are integers (the default)
                    fix_dataframe.rename(
                        columns={
                            c: "_" + str(c)
                            for c in fix_dataframe.columns  # pylint: disable=not-an-iterable
                        },
                        inplace=True,
                    )
                    exog = self.dm.join(fix_dataframe)
                # this handles the corner case where there technically is the
                # corresponding common effect, but the parameterization differs
                # between the common- and group-specific-effect specification. usually
                # this means the common effects use cell-means coding but the
                # group specific effects use k-1 coding
                else:
                    group = term.name.split("|")[1]
                    exog = self.model.group_specific_terms.values()
                    exog = [v.data.sum(1) for v in exog if v.name.split("|")[-1] == group]
                    index = ["_" + str(i) for i in range(len(exog))]
                    exog = pd.DataFrame(exog, index=index).T
                # this will replace self.mle (which is missing predictors)
                missing = "drop" if self.model.dropna else "none"
                full_mod = GLM(
                    endog=self.model.response.data,
                    exog=exog,
                    family=self.model.family.smfamily(self.model.family.smlink),
                    missing=missing,
                ).fit()
                sigma = self._get_slope_stats(
                    exog=exog, predictor=fix_data, full_mod=full_mod, sigma_corr=sigma_corr
                )

        # set the prior sigma.
        term.prior.args["sigma"].update(sigma=np.squeeze(np.atleast_1d(sigma)))

    def scale(self):
        # classify all terms
        common_intercepts = [
            term
            for term in self.model.terms.values()
            if not term.group_specific and term.data.sum(1).var() == 0
        ]
        common_slopes = [
            term
            for term in self.model.terms.values()
            if not term.group_specific and not term.data.sum(1).var() == 0
        ]
        group_specific_terms = [term for term in self.model.terms.values() if term.group_specific]

        # arrange them in the order in which they should be initialized
        term_list = common_slopes + common_intercepts + group_specific_terms
        term_types = (
            ["common"] * len(common_slopes)
            + ["intercept"] * len(common_intercepts)
            + ["group_specific"] * len(group_specific_terms)
        )

        # initialize them in order
        for term, term_type in zip(term_list, term_types):

            if term.prior.scale is None:
                # pylint: disable=protected-access
                if not term.prior._auto_scale or not self.model.auto_scale:
                    continue
                term.prior.scale = "wide"

            # MLE is fitted only once.
            if self.mle is None:
                self._fit_mle()

            # Convert scale names to float
            if isinstance(term.prior.scale, str):
                term.prior.scale = PriorScaler.names[term.prior.scale]

            # scale it!
            getattr(self, f"_scale_{term_type}")(term)

    def _fit_mle(self):
        """Fits MLE of the common part of the model.

        This used to be called in the class instantiation, but there is no need to fit the GLM when
        there are no automatic priors. So this method is only called when needed.
        """
        missing = "drop" if self.model.dropna else "none"
        try:
            self.mle = GLM(
                endog=self.model.response.data,
                exog=self.dm,
                family=self.model.family.smfamily(self.model.family.smlink),
                missing=missing,
            ).fit()
        except PerfectSeparationError as error:
            msg = "Perfect separation detected, automatic priors are not available. "
            msg += "Please indicate priors manually."
            raise PerfectSeparationError(msg) from error
        except:
            print("Unexpected error:", sys.exc_info()[0])
            raise



from abc import ABCMeta, abstractmethod


class BackEnd:
    """Base class for backend hierarchy."""

    __metaclass__ = ABCMeta

    @abstractmethod
    def build(self):
        pass

    @abstractmethod
    def run(self):
        pass



import logging

import numpy as np
import theano.tensor as tt
import pymc3 as pm

from bambi.priors import Prior
import bambi.version as version

from .base import BackEnd

_log = logging.getLogger("bambi")


def probit(x):
    """Probit function that ensures result is in (0, 1)"""
    eps = np.finfo(float).eps
    result = 0.5 + 0.5 * tt.erf(x / tt.sqrt(2))
    result = tt.switch(tt.eq(result, 0), eps, result)
    result = tt.switch(tt.eq(result, 1), 1 - eps, result)

    return result


def cloglog(x):
    """Cloglog function that ensures result is in (0, 1)"""
    eps = np.finfo(float).eps
    result = 1 - tt.exp(-tt.exp(x))
    result = tt.switch(tt.eq(result, 0), eps, result)
    result = tt.switch(tt.eq(result, 1), 1 - eps, result)

    return result


class PyMC3BackEnd(BackEnd):
    """PyMC3 model-fitting backend."""

    # Available link functions
    links = {
        "identity": lambda x: x,
        "logit": tt.nnet.sigmoid,
        "probit": probit,
        "cloglog": cloglog,
        "inverse": tt.inv,
        "inverse_squared": lambda x: tt.inv(tt.sqrt(x)),
        "log": tt.exp,
    }

    dists = {"HalfFlat": pm.Bound(pm.Flat, lower=0)}

    def __init__(self):
        self.name = pm.__name__
        self.version = pm.__version__

        # Attributes defined elsewhere
        self.model = None
        self.mu = None  # build()
        self.spec = None  # build()
        self.trace = None  # build()
        self.advi_params = None  # build()
        self.fit = False  # run()

    def build(self, spec):  # pylint: disable=arguments-differ
        """Compile the PyMC3 model from an abstract model specification.

        Parameters
        ----------
        spec : Bambi model
            A Bambi ``Model`` instance containing the abstract specification of the model
            to compile.
        """

        coords = spec._get_pymc_coords()  # pylint: disable=protected-access
        self.model = pm.Model(coords=coords)
        noncentered = spec.noncentered

        with self.model:
            self.mu = 0.0
            for term in spec.terms.values():
                data = term.data
                label = term.name
                dist_name = term.prior.name
                dist_args = term.prior.args
                dist_shape = term.data.shape[1]
                if dist_shape == 1:
                    dist_shape = ()
                coef = self._build_dist(
                    label, noncentered, dist_name, shape=dist_shape, **dist_args
                )

                if term.group_specific:
                    self.mu += coef[term.group_index][:, None] * term.predictor
                else:
                    self.mu += pm.math.dot(data, coef)[:, None]

            self._build_response(spec)
            self.spec = spec

    # pylint: disable=arguments-differ, inconsistent-return-statements
    def run(
        self, start=None, method="mcmc", init="auto", n_init=50000, omit_offsets=True, **kwargs
    ):
        """Run the PyMC3 MCMC sampler.

        Parameters
        ----------
        start: dict, or array of dict
            Starting parameter values to pass to sampler; see ``pm.sample()`` for details.
        method: str
            The method to use for fitting the model. By default, ``'mcmc'``, in which case the
            PyMC3 sampler will be used. Alternatively, ``'advi'``, in which case the model will be
            fitted using  automatic differentiation variational inference as implemented in PyMC3.
            Finally, ``'laplace'``, in which case a laplace approximation is used, ``'laplace'`` is
            not recommended other than for pedagogical use.
        init: str
            Initialization method (see PyMC3 sampler documentation). Currently, this is
            ``'jitter+adapt_diag'``, but this can change in the future.
        n_init: int
            Number of initialization iterations if ``init = 'advi'`` or '``init = 'nuts'``.
            Default is kind of in PyMC3 for the kinds of models we expect to see run with Bambi,
            so we lower it considerably.
        omit_offsets: bool
            Omits offset terms in the ``InferenceData`` object when the model includes
            group specific effects. Defaults to ``True``.

        Returns
        -------
        An ArviZ ``InferenceData`` instance.
        """
        model = self.model

        if method.lower() == "mcmc":
            draws = kwargs.pop("draws", 1000)
            with model:
                idata = pm.sample(
                    draws,
                    start=start,
                    init=init,
                    n_init=n_init,
                    return_inferencedata=True,
                    **kwargs,
                )

            if omit_offsets:
                offset_dims = [vn for vn in idata.posterior.dims if "offset" in vn]
                idata.posterior = idata.posterior.drop_dims(offset_dims)

            for group in idata.groups():
                getattr(idata, group).attrs["modeling_interface"] = "bambi"
                getattr(idata, group).attrs["modeling_interface_version"] = version.__version__

            self.fit = True
            return idata

        elif method.lower() == "advi":
            with model:
                self.advi_params = pm.variational.ADVI(start, **kwargs)
            return (
                self.advi_params
            )  # this should return an InferenceData object (once arviz adds support for VI)

        elif method.lower() == "laplace":
            return _laplace(model)

    def _build_dist(self, label, noncentered, dist, **kwargs):
        """Build and return a PyMC3 Distribution."""

        dist = self._get_dist(dist)
        kwargs = {k: self._expand_args(k, v, label, noncentered) for (k, v) in kwargs.items()}

        # Non-centered parameterization for hyperpriors
        if noncentered and has_hyperprior(kwargs):
            old_sigma = kwargs["sigma"]
            _offset = pm.Normal(label + "_offset", mu=0, sigma=1, shape=kwargs["shape"])
            return pm.Deterministic(label, _offset * old_sigma)

        return dist(label, **kwargs)

    def _get_dist(self, dist):
        """Return a PyMC3 distribution."""
        if isinstance(dist, str):
            if hasattr(pm, dist):
                dist = getattr(pm, dist)
            elif dist in self.dists:
                dist = self.dists[dist]
            else:
                raise ValueError(
                    f"The Distribution {dist} was not found in PyMC3 or the PyMC3BackEnd."
                )
        return dist

    def _build_response(self, spec):
        """Build and return a response distribution."""

        data = spec.response.data
        name = spec.response.name
        prior = spec.family.prior
        link = spec.family.link
        if isinstance(link, str):
            link = self.links[link]
        prior.args[spec.family.parent] = link(self.mu)
        prior.args["observed"] = data

        dist = self._get_dist(prior.name)
        kwargs = {k: self._expand_args(k, v, name, False) for (k, v) in prior.args.items()}

        if spec.family.name == "gamma":
            # Gamma distribution is specified using mu and sigma, but we request prior for alpha,
            # so we need to build sigma from mu and alpha.
            # we can just write kwargs["mu"] ** 2 / kwargs["alpha"]
            beta = kwargs["alpha"] / kwargs["mu"]
            sigma = (kwargs["mu"] / beta) ** 0.5
            return dist(name, mu=kwargs["mu"], sigma=sigma, observed=kwargs["observed"])

        return dist(name, **kwargs)

    def _expand_args(self, key, value, label, noncentered):
        # Inspect all args in case we have hyperparameters
        if isinstance(value, Prior):
            label = f"{label}_{key}"
            return self._build_dist(label, noncentered, value.name, **value.args)
        return value


def _laplace(model):
    """Fit a model using a Laplace approximation.

    Mainly for pedagogical use. ``mcmc`` and ``advi`` are better approximations.

    Parameters
    ----------
    model: PyMC3 model

    Returns
    -------
    Dictionary, the keys are the names of the variables and the values tuples of modes and standard
    deviations.
    """
    with model:
        varis = [v for v in model.unobserved_RVs if not pm.util.is_transformed_name(v.name)]
        maps = pm.find_MAP(start=model.test_point, vars=varis)
        hessian = pm.find_hessian(maps, vars=varis)
        if np.linalg.det(hessian) == 0:
            raise np.linalg.LinAlgError("Singular matrix. Use mcmc or advi method")
        stds = np.diag(np.linalg.inv(hessian) ** 0.5)
        maps = [v for (k, v) in maps.items() if not pm.util.is_transformed_name(k)]
        modes = [v.item() if v.size == 1 else v for v in maps]
        names = [v.name for v in varis]
        shapes = [np.atleast_1d(mode).shape for mode in modes]
        stds_reshaped = []
        idx0 = 0
        for shape in shapes:
            idx1 = idx0 + sum(shape)
            stds_reshaped.append(np.reshape(stds[idx0:idx1], shape))
            idx0 = idx1
    return dict(zip(names, zip(modes, stds_reshaped)))


def has_hyperprior(kwargs):
    """Determines if a Prior has an hyperprior"""

    return (
        "sigma" in kwargs
        and "observed" not in kwargs
        and isinstance(kwargs["sigma"], pm.model.TransformedRV)
    )



import codecs
import os
import sys

from setuptools import find_packages, setup

PROJECT_ROOT = os.path.dirname(os.path.realpath(__file__))
README_FILE = os.path.join(PROJECT_ROOT, "README.md")
VERSION_FILE = os.path.join(PROJECT_ROOT, "bambi", "version.py")
REQUIREMENTS_FILE = os.path.join(PROJECT_ROOT, "requirements.txt")


def get_long_description():
    with codecs.open(README_FILE, "rt") as buff:
        return buff.read()


def get_requirements():
    with codecs.open(REQUIREMENTS_FILE) as buff:
        return buff.read().splitlines()


with open(VERSION_FILE) as buff:
    exec(buff.read())

if len(set(("test", "easy_install")).intersection(sys.argv)) > 0:
    import setuptools

tests_require = []
extra_setuptools_args = {}
if "setuptools" in sys.modules:
    tests_require.append("nose")
    extra_setuptools_args = dict(
        test_suite="nose.collector", extras_require=dict(test="nose>=0.10.1")
    )

setup(
    name="bambi",
    version=__version__,
    description="BAyesian Model Building Interface in Python",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    url="http://github.com/bambinos/bambi",
    download_url="https://github.com/bambinos/bambi/archive/%s.tar.gz" % __version__,
    install_requires=get_requirements(),
    maintainer="Tal Yarkoni",
    maintainer_email="tyarkoni@gmail.com",
    packages=find_packages(exclude=["tests", "test_*"]),
    package_data={"bambi": ["config/*"]},
    tests_require=tests_require,
    license="MIT",
    **extra_setuptools_args,
)



from .priors import Family


def listify(obj):
    """Wrap all non-list or tuple objects in a list.

    Provides a simple way to accept flexible arguments.
    """
    if obj is None:
        return []
    else:
        return obj if isinstance(obj, (list, tuple, type(None))) else [obj]


def extract_family_prior(family, priors):
    # Given a family and a dictionary of priors, return the prior that could be
    # passed to update the prior of a parameter in that family.

    # Only built-in gaussian and negativebinomial have nuisance parameters
    # i.e, parameters not related to the link-transformed predicted outcome
    if isinstance(family, str):
        if family == "gaussian" and "sigma" in priors:
            return {"sigma": priors["sigma"]}
        elif family == "negativebinomial" and "alpha" in priors:
            return {"alpha": priors["alpha"]}
        elif family == "gamma" and "alpha" in priors:
            return {"alpha": priors["alpha"]}
    elif isinstance(family, Family):
        # Only work if there are nuisance parameters in the family, and if any of these nuisance
        # parameters is present in 'priors' dictionary.
        nuisance_params = [k for k in family.prior.args if k not in ["observed", family.parent]]
        if set(nuisance_params).intersection(set(priors)):
            return {k: priors[k] for k in nuisance_params if k in priors}
    return None


def link_match_family(link, family_name):  # pylint: disable= too-many-return-statements
    """Checks whether the a link can be used in a given family.

    When this function is used with built-in family names, it tests whether the link name can be
    used with the given built-in family. If the family name is not known, we return True because
    the user is working with a custom ``Family`` object.
    Which links can work with which families are taken from statsmodels.
    """
    if family_name == "gaussian":
        return link in ["identity", "log", "inverse"]
    elif family_name == "gamma":
        return link in ["identity", "log", "inverse"]
    elif family_name == "bernoulli":
        return link in ["identity", "logit", "probit", "cloglog"]
    elif family_name == "wald":
        return link in ["inverse", "inverse_squared", "identity", "log"]
    elif family_name == "negativebinomial":
        return link in ["identity", "log", "cloglog"]
    elif family_name == "poisson":
        return link in ["identity", "log"]
    else:  # Custom family, we don't know what link functions can be used
        return True



import itertools
import numpy as np


class ResponseTerm:
    """Representation of a single response model term.

    Parameters
    ----------
    term: formulae.ResponseVector
        An object describing the response of the model,
        as returned by ``formulae.design_matrices().response``
    prior : Prior
        A specification of the prior(s) to use. An instance of class ``priors.Prior``.
    family : str
        The name of the model family.
    """

    def __init__(self, term, prior=None, family=None):
        self.name = term.name
        self.data = term.design_vector
        self.categorical = term.type == "categoric"
        self.success_event = term.refclass if term.refclass is not None else 1
        self.prior = prior
        self.constant = np.var(self.data) == 0

        if family == "bernoulli":
            if not all(np.isin(self.data, ([0, 1]))):
                raise ValueError("Numeric response must be all 0 and 1 for 'bernoulli' family.")


class Term:
    """Representation of a single (common) model term.

    Parameters
    ----------
    name : str
        Name of the term.
    term_dict: dict
        A dictionary describing the components of the term. These can be found in
        ``formulae.design_matrices().common.terms_info``
    data : ndarray
        The term values.
    prior : Prior
        A specification of the prior(s) to use. An instance of class ``priors.Prior``.
    constant : bool
        Indicates whether the term levels collectively act as a constant, in which case the term is
        treated as an intercept for prior distribution purposes.
    """

    group_specific = False

    def __init__(self, name, term_dict, data, prior=None, constant=None):
        self.name = name
        self.data = data
        self.prior = prior
        self.categorical = term_dict["type"] == "categoric"
        self.levels = term_dict["full_names"]

        # identify and flag intercept and cell-means terms (i.e., full-rankdummy codes),
        # which receive special priors
        if constant is None:
            self.constant = np.atleast_2d(self.data.T).T.sum(1).var() == 0
        else:
            self.constant = constant

        if self.categorical:
            if "levels" in term_dict.keys():
                if term_dict["encoding"] == "full":
                    self.cleaned_levels = term_dict["levels"]
                else:
                    self.cleaned_levels = term_dict["levels"][1:]
            else:
                self.cleaned_levels = term_dict["reference"]
        else:
            self.cleaned_levels = None

        # Any interaction with 1 categorical is considered categorical (at least for now)
        if term_dict["type"] == "interaction":
            if any((v["type"] == "categoric" for v in term_dict["terms"].values())):
                self.categorical = True
                self.cleaned_levels = _interaction_labels(term_dict)


class GroupSpecificTerm:
    """Representation of a single (group specific) model term.

    Parameters
    ----------
    name : str
        Name of the term.
    term: dict
        A dictionary describing the components of the term. These can be found in
        ``formulae.design_matrices().group.terms_info``
    data : (DataFrame, Series, ndarray)
        The term values.
    prior : Prior
        A specification of the prior(s) to use. An instance of class ``priors.Prior``.
    constant : bool
        Indicates whether the term levels collectively act as a constant, in which case the term is
        treated as an intercept for prior distribution purposes.
    """

    group_specific = True

    def __init__(self, name, term, data, prior=None, constant=None):
        self.name = name
        self.data = data
        self.prior = prior
        self.categorical = term["type"] == "categoric"
        self.cleaned_levels = term["groups"]
        self.levels = term["full_names"]

        self.grouper = term["Ji"]
        self.predictor = term["Xi"]
        self.group_index = self.invert_dummies(self.grouper)

        if constant is None:
            self.constant = np.atleast_2d(self.data.T).T.sum(1).var() == 0
        else:
            self.constant = constant

    def invert_dummies(self, dummies):
        """
        For the sake of computational efficiency (i.e., to avoid lots of large matrix
        multiplications in the backends), invert the dummy-coding process and represent full-rank
        dummies as a vector of indices into the coefficients.
        """
        vec = np.zeros(len(dummies), dtype=int)
        for i in range(1, dummies.shape[1]):
            vec[dummies[:, i] == 1] = i
        return vec


def _interaction_labels(x):
    # taken from formulae
    terms = x["terms"]
    colnames = []

    for k, val in terms.items():
        if val["type"] in ["numeric"]:
            colnames.append([k])
        if val["type"] == "categoric":
            if "levels" in val.keys():
                if val["encoding"] == "full":
                    colnames.append([f"{k}[{level}]" for level in val["levels"]])
                else:
                    colnames.append([f"{k}[{level}]" for level in val["levels"][1:]])
            else:
                colnames.append([f"{k}[{val['reference']}]"])

    return [":".join(str_tuple) for str_tuple in list(itertools.product(*colnames))]
